import itertools
import warnings
from typing import Dict, List, Optional, Tuple, Union

import numpy as np
import torch
from opensr_test.config import Metric
from opensr_test.distance import DistanceMetric
from opensr_test.lightglue import DISK, LightGlue, SuperPoint
from opensr_test.lightglue.utils import rbd
from scipy.spatial.distance import cdist
from skimage.registration import phase_cross_correlation
from sklearn.linear_model import LinearRegression
from sklearn.pipeline import Pipeline, make_pipeline
from sklearn.preprocessing import PolynomialFeatures

# %-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-
# | Spatial transformation functions
# %-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-


def spatia_polynomial_fit(X: np.ndarray, y: np.ndarray, d: int) -> Pipeline:
    """Fit a polynomial regression using matched points

    Args:
        X (np.ndarray): Array with the x coordinates or y coordinates of the points (image 1)
        y (np.ndarray): Array with the x coordinates or y coordinates of the points (image 2)
        d (int): Degree of the polynomial

    Returns:
        Pipeline: The fitted model
    """

    pipe_model = make_pipeline(
        PolynomialFeatures(degree=d, include_bias=False), LinearRegression()
    ).fit(X, y)

    return pipe_model


def spatial_setup_model(
    features: str = "superpoint",
    matcher: str = "lightglue",
    max_num_keypoints: int = 2048,
    device: str = "cpu",
) -> tuple:
    """Setup the model for spatial check

    Args:
        features (str, optional): The feature extractor. Defaults to 'superpoint'.
        matcher (str, optional): The matcher. Defaults to 'lightglue'.
        max_num_keypoints (int, optional): The maximum number of keypoints. Defaults to 2048.
        device (str, optional): The device to use. Defaults to 'cpu'.

    Raises:
        ValueError: If the feature extractor or the matcher are not valid
        ValueError: If the device is not valid

    Returns:
        tuple: The feature extractor and the matcher models
    """

    # Local feature extractor
    if features == "superpoint":
        extractor = SuperPoint(max_num_keypoints=max_num_keypoints).eval().to(device)
    elif features == "disk":
        extractor = DISK(max_num_keypoints=max_num_keypoints).eval().to(device)
    else:
        raise ValueError(f"Unknown feature extractor {features}")

    # Local feature matcher
    if matcher == "lightglue":
        matcher = LightGlue(features=features).eval().to(device)
    else:
        raise ValueError(f"Unknown matcher {matcher}")

    return extractor, matcher


def spatial_get_matching_points(
    img01: torch.Tensor, img02: torch.Tensor, model: tuple, device: str = "cpu"
) -> Dict[str, torch.Tensor]:
    """Predict the spatial error between two images

    Args:
        img01 (torch.Tensor): A torch.tensor with the input image (B, H, W)
        img02 (torch.Tensor): A torch.tensor with the reference image (B, H, W)
        model (tuple): A tuple with the feature extractor and the matcher.
        device (str, optional): The device to use. Defaults to 'cpu'.

    Returns:
        Dict[str, torch.Tensor]: A dictionary with the points0, 
            points1, matches01.
    """

    # unpack the model - send to device
    extractor, matcher = model
    extractor = extractor.to(device)
    matcher = matcher.to(device)

    # Send the data to the device
    img01 = img01.to(device)
    img02 = img02.to(device)

    # extract local features
    with torch.no_grad():
        # auto-resize the image, disable with resize=None
        feats0 = extractor.extract(img01, resize=None)
        if feats0["keypoints"].shape[1] == 0:
            warnings.warn("No keypoints found in image 1")
            return False

        feats1 = extractor.extract(img02, resize=None)
        if feats1["keypoints"].shape[1] == 0:
            warnings.warn("No keypoints found in image 2")
            return False

        # match the features
        matches01 = matcher({"image0": feats0, "image1": feats1})

    # remove batch dimension
    feats0, feats1, matches01 = [rbd(x) for x in [feats0, feats1, matches01]]

    matches = matches01["matches"]
    points0 = feats0["keypoints"][matches[..., 0]]
    points1 = feats1["keypoints"][matches[..., 1]]

    matching_points = {"points0": points0, "points1": points1, "matches01": matches01}

    return matching_points


def spatial_model_fit(
    matching_points: Dict[str, torch.Tensor],
    n_points: Optional[int] = 10,
    threshold_distance: Optional[int] = 5,
    verbose: Optional[bool] = False,
) -> Union[np.ndarray, dict]:
    """Get a model that minimizes the spatial error between two images

    Args:
        matching_points (Dict[str, torch.Tensor]): A dictionary with the points0, 
            points1 and image size.
        n_points (Optional[int], optional): The minimum number of points. Defaults
            to 10.
        threshold_distance (Optional[int], optional): The maximum distance between
            the points. Defaults to 5 pixels.
        verbose (Optional[bool], optional): If True, print the error. Defaults to
            False.
        scale (Optional[int], optional): The scale factor to use. Defaults to 1.
        
    Returns:
        np.ndarray: The spatial error between the two images
    """

    points0 = matching_points["points0"]
    points1 = matching_points["points1"]

    # if the distance between the points is higher than 5 pixels,
    # it is considered a bad match
    dist = torch.sqrt(torch.sum((points0 - points1) ** 2, dim=1))
    thres = dist < threshold_distance
    p0 = points0[thres]
    p1 = points1[thres]

    # if not enough points, return 0
    if p0.shape[0] < n_points:
        warnings.warn("Not enough points to fit the model")
        return False

    # from torch.Tensor to numpy array
    p0 = p0.detach().cpu().numpy()
    p1 = p1.detach().cpu().numpy()

    # Fit a polynomial of degree 2 to the points
    X_img0 = p0[:, 0].reshape(-1, 1)
    X_img1 = p1[:, 0].reshape(-1, 1)
    model_x = spatia_polynomial_fit(X_img0, X_img1, 1)

    y_img0 = p0[:, 1].reshape(-1, 1)
    y_img1 = p1[:, 1].reshape(-1, 1)
    model_y = spatia_polynomial_fit(y_img0, y_img1, 1)

    # display error
    xoffset = np.round(model_x.predict(np.array(0).reshape(-1, 1)))
    yoffset = np.round(model_y.predict(np.array(0).reshape(-1, 1)))

    xhat = X_img0 + xoffset
    yhat = y_img0 + yoffset

    # full error
    full_error1 = np.sqrt((xhat - X_img1) ** 2 + (yhat - y_img1) ** 2)
    full_error2 = np.sqrt((X_img0 - X_img1) ** 2 + (y_img0 - y_img1) ** 2)

    if verbose:
        print(f"Initial [RMSE]: %.04f" % np.mean(full_error2))
        print(f"Final [RMSE]: %.04f" % np.mean(full_error1))

    to_export = {
        "offset": (int(xoffset), int(yoffset)),
        "error": (np.mean(full_error2), np.mean(full_error1)),
    }

    return to_export


def spatial_model_transform_pixel(
    image1: torch.Tensor, spatial_offset: tuple
) -> torch.Tensor:
    """ Transform the image according to the spatial offset obtained by the
    spatial_model_fit function. This correction is done at pixel level.

    Args:
        image1 (torch.Tensor): The image 1 with shape (B, H, W)
        spatial_offset (tuple): The spatial offset estimated by the 
            spatial_model_fit function.
    Returns:
        torch.Tensor: The transformed image
    """
    x_offs, y_offs = spatial_offset["offset"]

    # get max offset
    moffs = np.max(np.abs([x_offs, y_offs]))

    # Add padding according to the offset
    image_pad = torch.nn.functional.pad(
        image1, (moffs, moffs, moffs, moffs), mode="constant", value=0
    )

    if x_offs < 0:
        image_pad = image_pad[:, :, (moffs + x_offs) :]
    elif x_offs > 0:
        image_pad = image_pad[:, :, (moffs - x_offs) :]

    if y_offs < 0:
        image_pad = image_pad[:, (moffs - y_offs) :, :]
    elif y_offs > 0:
        image_pad = image_pad[:, (moffs + y_offs) :, :]

    # remove padding
    final_image = image_pad[:, 0 : image1.shape[1], 0 : image1.shape[2]]

    return final_image


def spatial_model_transform(
    lr_to_hr: torch.Tensor, hr: torch.Tensor, spatial_offset: tuple
) -> torch.Tensor:
    """ Transform the image according to the spatial offset

    Args:
        lr_to_hr (torch.Tensor): The low resolution image
        hr (torch.Tensor): The high resolution image
        spatial_offset (tuple): The spatial offset estimated by the 
            spatial_model_fit function.
    Returns:
        torch.Tensor: The transformed image
    """

    # Fix the image according to the spatial offset
    offset_image1 = spatial_model_transform_pixel(
        image1=lr_to_hr, spatial_offset=spatial_offset
    )
    hr_masked = hr * (offset_image1 != 0)

    # Create a mask with the offset image
    offset_image1 = offset_image1.detach().cpu().numpy()
    hr_masked = hr_masked.detach().cpu().numpy()

    # Subpixel refinement
    shift, error, diffphase = phase_cross_correlation(
        offset_image1.mean(0), hr_masked.mean(0), upsample_factor=100
    )

    # Fix the offset_image according to the subpixel refinement
    offset_image2 = spatial_model_transform_pixel(
        image1=torch.from_numpy(offset_image1).float(),
        spatial_offset={"offset": list(np.int16(np.round(shift)))},
    )

    return offset_image2


def spatial_aligment(
    sr: torch.Tensor,
    hr: torch.Tensor,
    spatial_model: tuple,
    threshold_n_points: Optional[int] = 5,
    threshold_distance: Optional[int] = 2 ** 63 - 1,
    rgb_bands: Optional[List[int]] = [0, 1, 2],
) -> torch.Tensor:
    """ Transform the image according to the spatial offset

    Args:
        sr (torch.Tensor): The super resolved image (B, C, H, W)
        hr (torch.Tensor): The high resolution image (B, C, H, W)
        spatial_model (tuple): A tuple with the feature extractor and the matcher models.
        threshold_n_points (Optional[int], optional): The minimum number of points to fit
            the linear model. Defaults to 5.
        threshold_distance (Optional[int], optional): The maximum distance 
            possible between the points. Defaults to 2**63 - 1.

    Returns:
        torch.Tensor: The transformed image (B, C, H, W).
    """
    # replace nan values with 0
    # lightglue does not work with nan values
    sr = torch.nan_to_num(sr, nan=0.0)
    hr = torch.nan_to_num(hr, nan=0.0)

    # get imag01 and img02
    if sr.shape[0] >= 3:
        img01 = sr[rgb_bands, ...]
        img02 = hr[rgb_bands, ...]
    else:
        img01 = sr[0][None]
        img02 = hr[0][None]

    # Get the matching points
    matching_points = spatial_get_matching_points(
        img01=img01, img02=img02, model=spatial_model
    )

    if matching_points is False:
        warnings.warn("Not enough points to align the images")
        return sr, matching_points

    # Fix the image according to the spatial offset
    spatial_offset = spatial_model_fit(
        matching_points=matching_points,
        n_points=threshold_n_points,
        threshold_distance=threshold_distance,
        verbose=False,
    )

    if spatial_offset is False:
        warnings.warn("Not enough valid points to align the images")
        return sr, matching_points

    # Fix the image according to the spatial offset
    offset_image = spatial_model_transform(
        lr_to_hr=sr, hr=hr, spatial_offset=spatial_offset
    )

    return offset_image, matching_points


# %-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-
# | Spatial transformation functions
# %-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-


def distance_matrix(
    x0: np.ndarray, y0: np.ndarray, x1: np.ndarray, y1: np.ndarray
) -> np.ndarray:
    """Calculate the distance matrix between two sets of points

    Args:
        x0 (np.ndarray): Array with the x coordinates of the points (image 1)
        y0 (np.ndarray): Array with the y coordinates of the points (image 1)
        x1 (np.ndarray): Array with the x coordinates of the points (image 2)
        y1 (np.ndarray): Array with the y coordinates of the points (image 2)

    Returns:
        np.ndarray: Array with the distances between the points
    """
    obs = np.vstack((x0, y0)).T
    interp = np.vstack((x1, y1)).T
    return cdist(obs, interp)


def linear_rbf(
    x: np.ndarray, y: np.ndarray, z: np.ndarray, xi: np.ndarray, yi: np.ndarray
) -> np.ndarray:
    """Interpolate using radial basis functions

    Args:
        x (np.ndarray): Array with the x coordinates of the points (image 1)
        y (np.ndarray): Array with the y coordinates of the points (image 1)
        z (np.ndarray): Array with the z coordinates of the points (image 1)
        xi (np.ndarray): Array with the x coordinates of the points (target image)
        yi (np.ndarray): Array with the y coordinates of the points (target image)
    
    Returns:
        np.ndarray: Array with the interpolated values
    """
    dist = distance_matrix(x, y, xi, yi)

    # Mutual pariwise distances between observations
    internal_dist = distance_matrix(x, y, x, y)

    # Now solve for the weights such that mistfit at the observations is minimized
    weights = np.linalg.solve(internal_dist, z)

    # Multiply the weights for each interpolated point by the distances
    zi = np.dot(dist.T, weights)

    return zi


def spatial_error(
    matching_points: Dict[str, torch.Tensor],
    threshold_distance: Optional[int] = 5,
    threshold_npoints: Optional[int] = 5,
    img_size: Optional[Tuple[int, int]] = (1024, 1024),
) -> Union[Tuple[float, Tuple[list, list]], Tuple[np.ndarray, Tuple[list, list]]]:
    """Calculate the spatial error between two images

    Args:
        matching_points (Dict[str, torch.Tensor]): A dictionary with the points0,
            points1 and image size.
        threshold_distance (Optional[int], optional): The maximum distance
            between the points. Defaults to 5 pixels.
        threshold_npoints (Optional[int], optional): The minimum number of 
            points. Defaults to 5.            
        degree (Optional[int], optional): The degree of the polynomial. Defaults to 1.
        img_size (Optional[Tuple[int, int]], optional): The size of the image. Defaults
            to (1024, 1024).
    Returns:
        Union[Tuple[float, Tuple[list, list]], Tuple[np.ndarray, Tuple[list, list]]] : 
            The spatial error between the two images. If grid=True, return the 
            interpolated grid using RBF. Also, return the points0 and points1 as 
            a tuple.
            
    Exceptions:
        ValueError: If not enough points to calculate the spatial error.
    """

    points0 = matching_points["points0"]
    points1 = matching_points["points1"]

    # if the distance between the points is higher than 5 pixels,
    # it is considered a bad match
    dist = torch.sqrt(torch.sum((points0 - points1) ** 2, dim=1))
    thres = dist < threshold_distance
    p0 = points0[thres]
    p1 = points1[thres]

    # if not enough points, return 0
    if p0.shape[0] < threshold_npoints:
        warnings.warn("Not enough points to calculate the spatial error")
        return np.array(np.nan)

    # from torch.Tensor to numpy array
    p0 = p0.detach().cpu().numpy()
    p1 = p1.detach().cpu().numpy()

    # Fit a polynomial of degree 2 to the points
    X_img0 = p0[:, 0].reshape(-1, 1)
    X_img1 = p1[:, 0].reshape(-1, 1)

    X_error = np.abs(X_img1 - X_img0)

    y_img0 = p0[:, 1].reshape(-1, 1)
    y_img1 = p1[:, 1].reshape(-1, 1)

    y_error = np.abs(y_img1 - y_img0)

    # Calculate the error
    full_error = np.sqrt(X_error ** 2 + y_error ** 2)

    # Interpolate the error
    arr_grid = np.array(
        list(
            itertools.product(
                np.arange(0, img_size[0], 1), np.arange(0, img_size[1], 1)
            )
        )
    )

    spgrid_1D = linear_rbf(
        X_img0.flatten(),
        y_img0.flatten(),
        full_error.flatten(),
        arr_grid[:, 0],
        arr_grid[:, 1],
    )

    spgrid_2D = np.flip(
        np.transpose(np.flip(spgrid_1D.reshape(img_size), axis=0), axes=(1, 0)), axis=1
    )

    return spgrid_2D


def spatial_metric(
    lr: torch.Tensor,
    sr_to_lr: torch.Tensor,
    spatial_model: tuple,
    threshold_n_points: Optional[int] = 5,
    threshold_distance: Optional[int] = 2 ** 63 - 1,
    description: str = "DISK Lightglue",
    device: str = "cpu",
) -> Union[float, np.ndarray]:
    """Calculate the spatial error between two images

    Args:
        lr (torch.Tensor): The LR image with shape (C, H, W)
        sr_to_lr (torch.Tensor): The SR degradated to the spatial resolution of
            LR with shape (C, H, W).
        spatial_model (tuple): A tuple with the feature extractor and the matcher models.
        threshold_distance (Optional[int], optional): The maximum distance between the
            points. Defaults to 5 pixels.
        threshold_npoints (Optional[int], optional): The minimum number of points.
            Defaults to 5.
        degree (Optional[int], optional): The degree of the polynomial. Defaults to 1.
        grid (Optional[bool], optional): If True, return the grid with the error. Defaults 
            to True.
        description (str, optional): The description of the metric. Defaults to 
            'DISK & Lightglue'.
        device (str, optional): The device to use. Defaults to 'cpu'.

    Returns:
        Union[float, np.ndarray]: The spatial error between the two images. If grid=True,
            return the grid with the error, otherwise return the mean error (RMSEmean).
    """

    # replace nan values with 0
    # lightglue does not work with nan values
    lr = torch.nan_to_num(lr, nan=0.0)
    sr_to_lr = torch.nan_to_num(sr_to_lr, nan=0.0)

    # Apply histogram matching
    image1 = lr.mean(0)[None]
    image2 = sr_to_lr.mean(0)[None]

    # Get the points and matches
    matching_points = spatial_get_matching_points(
        img01=image1, img02=image2, model=spatial_model, device=device
    )

    # is matching_points is a dict:
    if matching_points is False:
        return torch.tensor(torch.nan), matching_points

    # Calculate the error
    sp_error = spatial_error(
        matching_points=matching_points,
        threshold_distance=threshold_distance,
        threshold_npoints=threshold_n_points,
        img_size=image1[0].shape,
    )

    return torch.from_numpy(sp_error).float(), matching_points


class SpatialMetric(DistanceMetric):
    """Spectral information divergence between two tensors"""

    def __init__(
        self,
        x: torch.Tensor,
        y: torch.Tensor,
        name: str,
        threshold_n_points: int,
        threshold_distance: int,
        spatial_model: tuple,
        method: str = "image",
        patch_size: int = 32,
        device: str = "cpu",
    ):
        super().__init__(x=x, y=y, method=method, patch_size=patch_size, name=name)
        self.spatial_model = spatial_model
        self.threshold_n_points = threshold_n_points
        self.threshold_distance = threshold_distance
        self.device = device

        # Estimate the spatial error at pixel level
        self.metric, self.matching_points = spatial_metric(
            lr=x,
            sr_to_lr=y,
            spatial_model=self.spatial_model,
            threshold_n_points=self.threshold_n_points,
            threshold_distance=self.threshold_distance,
            device="cpu",
        )

    def _compute_image(self, x: torch.Tensor, y: torch.Tensor) -> torch.Tensor:
        return self.metric.mean()

    def compute_patch(self) -> torch.Tensor:
        metric_batched = self.do_square(self.metric[None], self.patch_size)
        metric_result = torch.zeros(metric_batched.shape[:2])
        xrange, yrange = metric_batched.shape[0:2]

        for x_index in range(xrange):
            for y_index in range(yrange):
                metric_batch = metric_batched[x_index, y_index]
                metric_result[x_index, y_index] = metric_batch.mean()

        return Metric(value=metric_result, description=self.name)

    def _compute_pixel(self, x: torch.Tensor, y: torch.Tensor) -> torch.Tensor:
        return self.metric



import torch
from opensr_test.distance import get_distance


def reflectance_metric(
    lr: torch.Tensor,
    sr_to_lr: torch.Tensor,
    metric: str,
    agg_method: str = "pixel",
    patch_size: int = 32,
):
    """ Calculate spectral metrics between two images.

    Args:
        lr (torch.Tensor): The LR image.
        sr_to_lr (torch.Tensor): The SR image converted to LR space.
        metric (str, optional): The metric to use. The following metrics 
            are supported: "sid", "sd", "pbias".
        grid (Optional[bool], optional): Whether to return the metric 
            as a grid or not
    """
    if not metric in ["kl", "l1", "l2", "pbias"]:
        raise ValueError(f"Invalid metric. Must be one of 'kl', 'l1', 'l2', 'pbias'")

    metric_value = get_distance(
        x=lr, y=sr_to_lr, method=metric, agg_method=agg_method, patch_size=patch_size
    ).compute()

    return metric_value



import io
import os
import re

from setuptools import find_packages, setup


def read(filename):
    filename = os.path.join(os.path.dirname(__file__), filename)
    text_type = type("")
    with io.open(filename, mode="r", encoding="utf-8") as fd:
        return re.sub(text_type(r":[a-z]+:`~?(.*?)`"), text_type(r"``\1``"), fd.read())


setup(
    name="opensr-test",
    version="0.0.999999",
    url="https://github.com/ESAOpenSR/opensr-test",
    license="MIT",
    author="Cesar Aybar Camacho",
    author_email="csaybar@gmail.com",
    description="A comprehensive benchmark for real-world Sentinel-2 imagery super-resolution",
    long_description=read("README.md"),
    long_description_content_type="text/markdown",
    packages=find_packages(
        exclude=("tests",), include=["opensr_test", "opensr_test.*"]
    ),
    install_requires=[
        "torch>=1.9.0",
        "numpy",
        "matplotlib",
        "scikit-image",
        "scikit-learn",
        "requests",
        "kornia",
        "pydantic",
        "opencv-python",
    ],
    extras_require={
        "perceptual": ["open_clip_torch", "lpips"]
    },
    classifiers=[
        "Development Status :: 2 - Pre-Alpha",
        "License :: OSI Approved :: MIT License",
        "Programming Language :: Python",
        "Programming Language :: Python :: 3.6",
        "Programming Language :: Python :: 3.7",
        "Programming Language :: Python :: 3.8",
        "Programming Language :: Python :: 3.9",
    ],
)



from typing import Dict, Optional, Tuple

import matplotlib.colors as colors
import matplotlib.pyplot as plt
import numpy as np
import torch
from opensr_test.lightglue import viz2d
from skimage import exposure


def min_max_range(tensor):
    """ 
    Return the min and max range of a tensor
    """
    tensor[torch.isnan(tensor)] = 0
    rmax = torch.tensor([tensor.min(), tensor.max()]).abs().min()
    rmin = rmax * -1
    return rmin, rmax


def linear_fix(img: torch.Tensor, permute=True) -> torch.Tensor:
    """ Linearly stretch the values of the image to increase contrast.

    Args:
        img (torch.Tensor): The RGB image to stretch (3xHxW).
        permute (bool, optional): Permute the dimensions to HxWx3.
            Defaults to True.
    Returns:
        torch.Tensor: The stretched image (HxWx3)
    """
    img = img.detach().cpu().numpy()
    container = []
    for i in range(3):
        p2, p98 = np.percentile(img[i, ...], (2, 98))
        img_rescale = exposure.rescale_intensity(img[i, ...], in_range=(p2, p98))
        container.append(img_rescale)
    new_img = np.array(container)
    if permute:
        new_img = np.moveaxis(new_img, 0, -1)
    return torch.from_numpy(new_img).float()


def do_nothing(img: torch.Tensor, permute=True) -> torch.Tensor:
    """ Do nothing to the image.
    
    Args:
        img (torch.Tensor): The RGB image to stretch (3xHxW).
        permute (bool, optional): Permute the dimensions to HxWx3.
            Defaults to True.
    Returns:
        torch.Tensor: The stretched image (HxWx3)
    """
    img = img.detach().cpu().numpy()
    container = []
    for i in range(3):
        img_rescale = img[i, ...]
        container.append(img_rescale)
    new_img = np.array(container)
    if permute:
        new_img = np.moveaxis(new_img, 0, -1)
    return torch.from_numpy(new_img).float()


def equalize_hist(img: torch.Tensor, permute=True) -> torch.Tensor:
    """ Equalize the histogram of the image to increase contrast.

    Args:
        img (torch.Tensor): The RGB image to stretch (3xHxW).
        permute (bool, optional): Permute the dimensions to HxWx3.
    Returns:
        torch.Tensor: The stretched image (HxWx3)
    """
    img = img.detach().cpu().numpy()
    container = []
    for i in range(3):
        img_rescale = exposure.equalize_hist(img[i, ...])
        container.append(img_rescale)
    new_img = np.array(container)
    if permute:
        new_img = np.moveaxis(new_img, 0, -1)
    return torch.from_numpy(new_img).float()


def quadruplets(
    lr_img: torch.Tensor,
    sr_img: torch.Tensor,
    hr_img: torch.Tensor,
    landuse_img: Optional[torch.Tensor] = None,
    stretch: Optional[str] = "linear",
) -> Tuple[plt.figure, plt.Axes]:
    """ Display LR, SR, HR and Landuse images in a single figure.

    Args:
        lr_img (torch.Tensor): The LR RGB (3xHxW) image.
        sr_img (torch.Tensor): The SR RGB (3xHxW) image.
        hr_img (torch.Tensor): The HR RGB (3xHxW) image.
        landuse_img (torch.Tensor): The landuse grayscale (HxW) image.
            Optional, defaults to None.
        stretch (Optional[str], optional): Option to stretch the values 
            to increase contrast: "lin" (linear) or "hist" (histogram)

    Raises:
        ValueError: The LR image must be a RGB (3xHxW) image.
        ValueError: The SR image must be a RGB (3xHxW) image.
        ValueError: The HR image must be a RGB (3xHxW) image.
        ValueError: The landuse image must be a grayscale (HxW) image.

    Returns:
        Tuple[plt.figure, plt.Axes]: The figure and axes to plot.
    """
    if lr_img.shape[0] == 4:
        raise ValueError("The LR image must be a RGB (3xHxW) image.")
    if sr_img.shape[0] == 4:
        raise ValueError("The SR image must be a RGB (3xHxW) image.")
    if hr_img.shape[0] == 4:
        raise ValueError("The HR image must be a RGB (3xHxW) image.")

    if landuse_img is None:
        return triplets(lr_img, sr_img, hr_img, stretch=stretch)

    # Apply the stretch
    if stretch == "linear":
        lr_img = linear_fix(lr_img)
        sr_img = linear_fix(sr_img)
        hr_img = linear_fix(hr_img)
    elif stretch == "histogram":
        lr_img = equalize_hist(lr_img)
        sr_img = equalize_hist(sr_img)
        hr_img = equalize_hist(hr_img)

    # Define the categorical values and colors
    values = [10, 20, 30, 40, 50, 60, 70, 80, 90, 95, 100]
    colors_list = [
        "#006400",
        "#ffbb22",
        "#ffff4c",
        "#f096ff",
        "#fa0000",
        "#b4b4b4",
        "#f0f0f0",
        "#0064c8",
        "#0096a0",
        "#00cf75",
        "#fae6a0",
    ]
    labels = [
        "Tree cover (10)",
        "Shrubland (20)",
        "Grassland (30)",
        "Cropland (40)",
        "Built-up (50)",
        "Bare / sparse vegetation (60)",
        "Snow and ice (70)",
        "Permanent water bodies (80)",
        "Herbaceous wetland (90)",
        "Mangroves (95)",
        "Moss and lichen (100)",
    ]

    # Set bounds and ticks
    bounds = [values[i] - 2.5 for i in range(len(values))]
    bounds.append(values[-1] + 2.5)
    ticks = [values[i] for i in range(len(values))]

    # Create a iterable colormap
    cmap = colors.ListedColormap(colors_list)
    norm = colors.BoundaryNorm(bounds, cmap.N)
    sm = plt.cm.ScalarMappable(cmap=cmap, norm=norm)
    sm.set_array([])

    # Plot the images
    scale_factor = hr_img.shape[0] / lr_img.shape[0]
    fig, axs = plt.subplots(1, 4, figsize=(25, 5))
    axs[0].imshow(lr_img)
    axs[0].set_title("LR")
    axs[1].imshow(sr_img)
    axs[1].set_title("SR")
    axs[2].imshow(hr_img)
    axs[2].set_title("HR")
    axs[3].imshow(landuse_img, cmap=cmap, interpolation="nearest", norm=norm)
    axs[3].set_title("LandUse")

    # Add the colorbar
    cbar_ax = axs[0].inset_axes([1.05, 0, 0.05, 1], transform=axs[3].transAxes)
    cbar = fig.colorbar(sm, cax=cbar_ax, ticks=ticks, boundaries=bounds)
    cbar.ax.set_yticklabels(labels)

    # Add the suptitle
    fig.suptitle("Scale factor: %.2f" % scale_factor, fontsize=16)

    # return the figure to plot
    return fig, axs


def triplets(
    lr_img: torch.Tensor,
    sr_img: torch.Tensor,
    hr_img: torch.Tensor,
    stretch: Optional[str] = "linear",
) -> Tuple[plt.figure, plt.Axes]:
    """ Display LR, SR, HR and Landuse images in a single figure.

    Args:
        lr_img (torch.Tensor): The LR RGB (3xHxW) image.
        sr_img (torch.Tensor): The SR RGB (3xHxW) image.
        hr_img (torch.Tensor): The HR RGB (3xHxW) image.
        stretch (Optional[str], optional): Option to stretch the values
            to increase contrast: "lin" (linear) or "hist" (histogram)

    Raises:
        ValueError: The LR image must be a RGB (3xHxW) image.
        ValueError: The SR image must be a RGB (3xHxW) image.
        ValueError: The HR image must be a RGB (3xHxW) image.
        ValueError: The landuse image must be a grayscale (HxW) image.

    Returns:
        Tuple[plt.figure, plt.Axes]: The figure and axes to plot.
    """
    if lr_img.shape[0] == 4:
        raise ValueError("The LR image must be a RGB (3xHxW) image.")
    if sr_img.shape[0] == 4:
        raise ValueError("The SR image must be a RGB (3xHxW) image.")
    if hr_img.shape[0] == 4:
        raise ValueError("The HR image must be a RGB (3xHxW) image.")

    # Apply the stretch
    if stretch == "linear":
        lr_img = linear_fix(lr_img)
        sr_img = linear_fix(sr_img)
        hr_img = linear_fix(hr_img)
    elif stretch == "histogram":
        lr_img = equalize_hist(lr_img)
        sr_img = equalize_hist(sr_img)
        hr_img = equalize_hist(hr_img)
    else:
        lr_img = do_nothing(lr_img)
        sr_img = do_nothing(sr_img)
        hr_img = do_nothing(hr_img)

    # Plot the images
    scale_factor = hr_img.shape[0] / lr_img.shape[0]
    fig, axs = plt.subplots(1, 3, figsize=(25, 5))
    axs[0].imshow(lr_img)
    axs[0].set_title("LR")
    axs[1].imshow(sr_img)
    axs[1].set_title("SR")
    axs[2].imshow(hr_img)
    axs[2].set_title("HR")

    # Add the suptitle
    fig.suptitle("Scale factor: %.2f" % scale_factor, fontsize=16)
    
    # return the figure to plot
    return fig, axs


def spatial_matches(
    lr: torch.Tensor,
    sr_to_lr: torch.Tensor,
    points0: torch.Tensor,
    points1: torch.Tensor,
    matches01: Dict[str, torch.Tensor],
    threshold_distance: Optional[int] = 5,
    stretch: Optional[str] = "linear",
):

    if stretch == "linear":
        sr_to_lr = linear_fix(sr_to_lr, permute=False)
        lr = linear_fix(lr, permute=False)
    elif stretch == "histogram":
        sr_to_lr = equalize_hist(sr_to_lr, permute=False)
        lr = equalize_hist(lr, permute=False)

    # if the distance between the points is higher than
    # threshold_distance pixels, it is considered a bad match
    dist = torch.sqrt(torch.sum((points0 - points1) ** 2, dim=1))
    thres = dist < threshold_distance
    p0 = points0[thres]
    p1 = points1[thres]

    axes = viz2d.plot_images([lr, sr_to_lr], titles=["LR", "SR to LR"])
    viz2d.plot_matches(p0, p1, color="lime", lw=0.2)
    viz2d.add_text(0, f'Stop after {matches01["stop"]} layers', fs=20)
    return axes


def display_results(
    lr: torch.Tensor,
    lrdown: torch.Tensor,
    sr: torch.Tensor,
    srharm: torch.Tensor,
    hr: torch.Tensor,
    e1: torch.Tensor,
    e1_title: str,
    e1_subtitle: str,
    e2: torch.Tensor,
    e2_title: str,
    e2_subtitle: str,
    e3: torch.Tensor,
    e3_points: Tuple[torch.Tensor, torch.Tensor],
    e3_title: str,
    e3_subtitle: str,
    e4: torch.Tensor,
    e4_title: str,
    e4_subtitle: str,
    e5: torch.Tensor,
    e5_title: str,
    e5_subtitle: str,
    stretch: Optional[str] = "linear",
):
    """ Display the results of the SR algorithm

    Args:
        lr (torch.Tensor): The LR image (3, H, W).
        sr (torch.Tensor): The SR image (3, H, W).
        hr (torch.Tensor): The HR image (3, H, W).
        e1 (torch.Tensor): The local reflectance map error (H, W).
        e1_title (str): The local reflectance error method.
        e1_subtitle (str): The median value of the local reflectance error.
        e2 (torch.Tensor): The spatial local errors (H, W).
        e2_points (Tuple[torch.Tensor, torch.Tensor]): The points to plot on 
            the spatial local errors.
        e2_title (str): The spatial local error method.
        e2_subtitle (str): The median value of the spatial local error.
        e3 (torch.Tensor): The HF map error (H, W).
        e3_title (str): The HF map error method.
        e3_subtitle (str): The median value of the HF map error.
        e4 (torch.Tensor): The improvement ratio (H, W).
        e4_title (str): The improvement ratio method.
        e4_subtitle (str): The median value of the improvement ratio.
        e5 (torch.Tensor): The hallucination error (H, W).
        e5_title (str): The hallucination error method.
        e5_subtitle (str): The median value of the hallucination error.

    Returns:
        fig, axs: The figure and axes of the plot.
    """

    # Apply the stretch
    if stretch == "linear":
        lr = linear_fix(lr)
        lrdown = linear_fix(lrdown)
        sr = linear_fix(sr)
        srharm = linear_fix(srharm)
        hr = linear_fix(hr)

    elif stretch == "histogram":
        lr = equalize_hist(lr)
        lrdown = equalize_hist(lrdown)
        sr = equalize_hist(sr)
        srharm = equalize_hist(srharm)
        hr = equalize_hist(hr)
    else:
        lr = do_nothing(lr)
        lrdown = do_nothing(lrdown)
        sr = do_nothing(sr)
        srharm = do_nothing(srharm)
        hr = do_nothing(hr)

    # Create the figure and axes (remove white space around the images)
    fig, axs = plt.subplots(2, 5, figsize=(20, 10), tight_layout=True)

    # Desactive all the axis
    for ax in axs.flatten():
        ax.axis("off")

    # Plot the first row (lr, sr, hr)
    axs[0, 0].imshow(lr)
    axs[0, 0].set_title("LR", fontsize=20, fontweight="bold")
    axs[0, 1].imshow(lrdown)
    axs[0, 1].set_title("LRdown", fontsize=20, fontweight="bold")
    axs[0, 2].imshow(sr)
    axs[0, 2].set_title("SR", fontsize=20, fontweight="bold")
    axs[0, 3].imshow(srharm)
    axs[0, 3].set_title("SRharm", fontsize=20, fontweight="bold")
    axs[0, 4].imshow(hr)
    axs[0, 4].set_title("HR", fontsize=20, fontweight="bold")

    # Display the local reflectance map error
    axs[1, 0].imshow(e1)
    axs[1, 0].set_title(
        "%s \n %s: %s" % (r"$\bf{Reflectance\ Consistency \downarrow}$", e1_title, e1_subtitle)
    )

    # Display the spectral map error
    axs[1, 1].imshow(e2)
    axs[1, 1].set_title(
        "%s \n %s: %s" % (r"$\bf{Spectral\ Consistency \downarrow}$", e2_title, e2_subtitle)
    )

    # Display the Spatial consistency
    if len(torch.tensor(e3).shape) == 0:
        axs[1, 2].imshow(e2 * torch.nan)
    else:
        axs[1, 2].imshow(e3, vmin=e3.min(), vmax=e3.max(), cmap="RdBu")
        axs[1, 2].plot(
            [x[0] for x in e3_points], [x[1] for x in e3_points], "r*", markersize=5
        )
    axs[1, 2].set_title(
        "%s \n %s: %s" % (r"$\bf{Spatial\ Consistency \downarrow}$", e3_title, e3_subtitle)
    )

    # Display the distance to the ommission space
    axs[1, 3].imshow(e4)
    axs[1, 3].set_title(
        "%s \n %s: %s"
        % (r"$\bf{Distance\ to\ Omission\ Space \uparrow}$", e4_title, e4_subtitle)
    )

    # Display the Hallucination error
    axs[1, 4].imshow(e5)
    axs[1, 4].set_title(
        "%s \n %s: %s"
        % (r"$\bf{Distance\ to\ Improvement\ Space \downarrow}$", e5_title, e5_subtitle)
    )

    return fig, axs


def display_tc_score(
    sr_rgb: torch.Tensor,
    d_im_ref: torch.Tensor,
    d_om_ref: torch.Tensor,
    tc_score: torch.Tensor,
    log_scale: bool = True,
    stretch: Optional[str] = "linear",
):    
    # Apply the stretch
    if stretch == "linear":
        sr_rgb = linear_fix(sr_rgb)
    elif stretch == "histogram":
        sr_rgb = equalize_hist(sr_rgb)
    else:
        sr_rgb = do_nothing(sr_rgb)

    # Custom categorical colormap - Blue[0], Green[1], Red[2]
    colorbar = colors.ListedColormap(["blue", "green", "red"])
    
    # Define triplets
    p1 = d_im_ref.ravel()
    p1 = p1[~torch.isnan(p1)]

    p2 = d_om_ref.ravel()
    p2 = p2[~torch.isnan(p2)]
    
    p3 = tc_score.ravel()
    p3 = p3[~torch.isnan(p3)]
    
    if log_scale:
        fig, ax = plt.subplots(1, 3, figsize=(15, 5))
        ax[0].imshow(sr_rgb)
        ax[0].set_title("SR RGB", fontsize=20, fontweight="bold")
        ax[1].imshow(tc_score, cmap=colorbar)
        ax[1].set_title("TC score - GRID", fontsize=20, fontweight="bold")
        ax[2].scatter(p1, p2, c=p3, cmap=colorbar)
        ax[2].set_ylabel("$d_{im}$", fontsize=18)
        ax[2].set_xlabel("$d_{om}$", fontsize=18)
        ax[2].set_title("TC score - 2D", fontsize=20, fontweight="bold")
        ax[2].set_yscale("log")
        ax[2].set_xscale("log")    
        # make square and equal
        ax[2].set_aspect(1.0/ax[2].get_data_ratio(), adjustable='box')
    else:
        fig, ax = plt.subplots(1, 3, figsize=(15, 5))
        ax[0].imshow(sr_rgb.permute(1, 2, 0)*3)
        ax[0].set_title("SR RGB", fontsize=20, fontweight="bold")
        ax[1].imshow(tc_score, cmap=colorbar)
        ax[1].set_title("TC score - GRID", fontsize=20, fontweight="bold")
        ax[2].scatter(p1, p2, c=p3, cmap=colorbar)        
        ax[2].set_ylabel("$d_{im}$", fontsize=18)
        ax[2].set_xlabel("$d_{om}$", fontsize=18)
        ax[2].set_title("TC score - 2D", fontsize=20, fontweight="bold")        
    return fig, ax
